;;; custom-load.el --- automatically extracted custom dependencies

;;; Code:

(autoload 'custom-add-loads "cus-load")

(custom-add-loads 'mercurial '("mercurial"))
(custom-add-loads 'tools '("mercurial" "vc"))
(custom-add-loads 'pcl-cvs '("log-edit"))
(custom-add-loads 'backup '("vc-hooks"))
(custom-add-loads 'log-edit '("log-edit"))
(custom-add-loads 'vc '("log-edit" "vc-arch" "vc-cvs" "vc-hooks" "vc-mcvs" "vc-rcs" "vc-sccs" "vc-svn" "vc"))

;;; custom-load.el ends here
